function setup_screen()
  love.window.setTitle('Les Poissons Danse')  -- ウィンドウタイトル
  width  = 480                       -- ゲーム画面の横幅
  height = 360                       -- ゲーム画面の高さ
  swidth, sheight, mode = love.window.getMode()  --  ウィンドウの横幅・高さ・モード
  mode.vsync = true
  love.window.setMode(swidth, sheight, mode)
  if swidth / sheight > width / height then -- 実画面の方が横長
    scale = sheight / height          -- 縦方向で倍率を決める
    transx = math.floor((swidth - width * scale) / 2)  -- 空白の幅
    if transx < 0 then transx = 0 end
    transy = 0
  else
    scale = swidth / width            -- 横方向で倍率を決める
    transx = 0
    transy = math.floor((sheight - height * scale) / 2) -- 空白の高さ
    if transy < 0 then transy = 0 end
  end
  love.graphics.setBackgroundColor(0, 0, 60)  -- 背景を濃い青色に
  msg_height = 0                      -- メッセージエリアの高さ
  font = love.graphics.newFont(16)    -- スコア表示用のフォント
end

function love.load()
  setup_screen()  -- 画面の初期化
  -- 画像の読み込み
  images = {}
  images.fish0 = love.graphics.newImage('images/fish0.png') -- さかな
  images.man0 = love.graphics.newImage('images/man0.png')   -- 人１
  images.man2 = love.graphics.newImage('images/man2.png')   -- 人２
  images.dai0 = love.graphics.newImage('images/dai0.png')   -- 眠り
  images.dai1 = love.graphics.newImage('images/dai1.png')   -- 笑い
  images.stone = love.graphics.newImage('images/stone.png') -- 石
  images.title1 = love.graphics.newImage('images/title1.png')  -- 'Les Poissons Danse'
  images.title2 = love.graphics.newImage('images/title2.png')  -- 'さかなは踊る'
  images.title3 = love.graphics.newImage('images/title3.png')  -- 'Press START Button'
  images.title4 = love.graphics.newImage('images/title4.png')  -- 'Hit SPACE Bar'
  images.clear1 = love.graphics.newImage('images/clear1.png')  -- 'This Stage Clear'
  -- サウンドの読み込み
  -- あまり長いものはないので、すべて static で読み込む
  local i
  sounds = {}
  sounds.bgm = {}
  for i = 1, 12 do
    sounds.bgm[i] = love.audio.newSource(string.format("sounds/bgm%02d.mp3", i), "static")
  end
  sounds.title = love.audio.newSource("sounds/bgm_title.mp3", "static")
  sounds.start = love.audio.newSource("sounds/bgm_start.mp3", "static")
  sounds.clear = love.audio.newSource("sounds/bgm_clear.mp3", "static")
  sounds.pomp = love.audio.newSource("sounds/bgm_pomp.mp3", "static")
  sounds.catch = love.audio.newSource("sounds/bgm_catch.mp3", "static")
  sounds.dead = love.audio.newSource("sounds/bgm_dead.mp3", "static")
  sounds.title:setLooping(true)
  -- 方向テーブル
  local s2 = 0.707107
  direct = {{1, 0}, {s2, s2}, {0, 1}, {-s2, s2}, {-1, 0}, {-s2, -s2}, {0, -1}, {s2, -s2}}
  rdirect = {5, 4, 3, 6, -1, 2, 7, 0, 1}
  rotate = {0, 1, 1, -1, 0, 0, 1, 0}
  invert = {-1, -1, -1, 1, 1, 1, 1, -1}
  -- status: 0=タイトル、1=ゲームスタート、2=ステージスタート、3=ゲーム中、
  -- 4=アトラクション、5=フェードアウト、6=ゲームオーバー
  S_TITLE, S_NEWGAME, S_NEWSTAGE, S_GAME, S_SHOW, S_FADEOUT, S_OVER = 0, 1, 2, 3, 4, 5, 6
  change_status(S_TITLE)
  fish = nil
end

-- 状態を変える
function change_status(n)
  status = n
  stime = nil
end

-- 新しいステージ
function prepare_stage()
  nfish = 8
  nstones = 0
  fish = {}
  man = {}
  nghosts = stage
  ghosts = {}
  setup_man()
  setup_fish()
  setup_ghost()
  stage_time = 120  --  １ステージの時間
  rem_time = stage_time --  残り時間
  --  BGM のカウンタ：ワンフレーズ演奏するごとに１増やす
  --  演奏するフレーズ：sounds.bgm[n], n = (mc < 42 ? 1 : (mc - 37) div 5)
  --  演奏時間：240 / (180 + (n - 1) * 8)秒
  mc = 0
  mtime = nil -- 次のフレーズを開始する時刻
end

-- オブジェクト a から b は距離 r 以内にあるか？
-- もしあれば、a から見た b の方向を返す (0-7)。なければ -1 を返す
function collide(a, b, r)
  local x = b.x - a.x
  local y = b.y - a.y
  if x * x + y * y < r * r then
    local rad = math.floor(math.atan2(y, x) * 4 / 3.1416)
    return (rad + 8) % 8
  else
    return -1
  end
end

-- if 文の簡略化：f が真なら a, 偽なら b を返す
function sel(f, a, b)
  if f then return a else return b end
end

-- ヒトを発生させる（画面中央）
function setup_man()
  local wid
  wid = images.man0:getWidth()
  man.x = math.floor((width - wid) / 2)
  man.y = math.floor((height - msg_height - wid) / 2) + msg_height
  man.alive = 1
  man.direction = 0
  man.speed = 150
  man.kind = 0 -- ヒト
  man.img = images.man0
  man.hw = man.img:getWidth() / 2
end

-- さかなを発生させる
function setup_fish()
  local i, j, wid
  fish = {}
  alive = nfish
  wid = images.fish0:getWidth()
  for i = 1, nfish do
    local f = {}
    repeat
      f.x = math.random() * (width - wid) + wid / 2
      f.y = math.random() * (height - msg_height - wid) + wid / 2 + msg_height
      f.alive = true
      if man and man.alive and collide(f, man, wid * 3) >= 0 then
        f.alive = false
      else
        for j = 1, i - 1 do
          if collide(f, fish[j], wid * 2) >= 0 then
            f.alive = false
            break
          end
        end
      end
    until f.alive
    f.speed = 100
    f.direction = math.floor(math.random() * 8)
    f.kind = 1 -- さかな
    f.img = images.fish0
    f.hw = f.img:getWidth() / 2
    fish[i] = f
  end
end

-- ゴーストを発生させる
function setup_ghost()
  local i, j, wid
  wid = images.dai0:getWidth()
  for i = 1, nghosts do
    local g = {}
    repeat
      g.x = math.random() * (width - wid) + wid / 2
      g.y = math.random() * (height - msg_height - wid) + wid / 2 + msg_height
      g.alive = true
      if man and man.alive and collide(g, man, wid * 4) >= 0 then
        g.alive = false
      else
        for j = 1, nfish + nghosts do
          local f
          if j <= nfish then
            f = fish[j]
          else
            f = ghosts[j - nfish]
          end
          if f and collide(g, f, wid * 2) >= 0 then
            g.alive = false
            break
          end
        end
      end
    until g.alive
    g.speed = 40
    g.direction = math.floor(math.random() * 8)
    g.kind = 2 -- ゴースト
    g.img = images.dai0
    g.hw = g.img:getWidth() / 2
    g.life = math.floor(math.random() * 10) + 10
    ghosts[i] = g
  end
end

-- r の反対±1 の方向を返す
function opposite(r)
  return (r + 3 + math.floor(math.random() * 3)) % 8
end

function movexy(x, y, dir, ds)
  local xx = x + direct[dir + 1][1] * ds
  local yy = y + direct[dir + 1][2] * ds
  return xx, yy
end

-- キャラクタを時間 dt 分だけ進める
function proceed(a, dt)
  local x, y, xsave, ysave, r, f, x1, y1, i, limit, dir, tbl1, xmin, xmax, ymin, ymax, ds
  if a.dead or a.hidden then return end
  -- 座標などをローカル変数に取り出す
  xsave, ysave, dir = a.x, a.y, a.direction
  ds = a.speed * dt
  -- x, y の可能な範囲
  xmin = a.hw
  xmax = width - a.hw
  ymin = msg_height + a.hw
  ymax = height - a.hw
  -- dt 分だけ動かしてみる
  x, y = movexy(xsave, ysave, dir, ds)
  -- 壁にぶつかっているか
  r = 4
  if x < xmin then
    r = 1
  elseif x >= xmax then
    r = 7
  end
  if y < ymin then
    r = r - 1
  elseif y >= ymax then
    r = r + 1
  end
  if r == 0 then
    dir = 1
  elseif r == 1 then
    dir = opposite(4)
  elseif r == 2 then
    dir = 7
  elseif r == 3 then
    dir = opposite(6)
  elseif r == 5 then
    dir = opposite(2)
  elseif r == 6 then
    dir = 3
  elseif r == 7 then
    dir = opposite(0)
  elseif r == 8 then
    dir = 5
  end
  if dir ~= a.direction then
    x, y = movexy(xsave, ysave, dir, ds)  --  計算し直す
  end
  -- 仮に座標を設定
  a.x, a.y = x, y
  -- 衝突判定
  if a.kind == 0 then -- 人
    for i = 1, nfish do
      f = fish[i]
      r = collide(a, f, 32)
      if r >= 0 then  -- さかなをつかまえた
        if not f.dead then  -- 生きているか？
          f.dead = true     -- 石になる
          f.direction = 0
          f.img = images.stone
          nstones = nstones + 1
          score = score + (stage + 1) * 5
          sounds.catch:play()
          if nstones == nfish then
            -- ステージクリア
            love.audio.stop()  -- BGMを全部止める
            draw_present()     -- 画面を更新
            sounds.clear:play()
            wait_duration(sounds.clear:getDuration()) -- 鳴り終わるまで待つ
            score = score + rem_time * (stage + 1) * 5
            if stage % 4 == 0 then
              change_status(S_SHOW)
            else
              stage = stage + 1
              change_status(S_NEWSTAGE)
            end
            return
          end
        end
        x, y = xsave, ysave -- 立ち止まる
        break
      end
    end
    a.x, a.y = x, y  -- 座標をアップデート
  elseif a.kind == 1 then -- さかな
    if man then
      r = collide(a, man, 80)
      if r >= 0 and (a.direction - r + 9) % 8 <= 2 then  -- 進む方向に人がいる
        dir = opposite(r)          -- 遠ざかる
      end
    end
    if dir == a.direction then
      for i = 1, nfish + nghosts do
        if i <= nfish then f = fish[i] else f = ghosts[i - nfish] end
        if f and f ~= a then -- 自分以外のさかな、ゴースト
          r = collide(a, f, 32)
          if r >= 0 then
            dir = opposite(dir)  -- 向きを変える
            break
          end
        end
      end
    end
  elseif a.kind == 2 then -- ゴースト
    r = collide(a, man, 160)
    if r >= 0 then  -- 近くに人がいる
      dir = opposite(r + 4)  -- 近づく
      a.laugh = true   -- にやける
      a.img = images.dai1
      if collide(a, man, 32) >= 0 then
        change_status(S_FADEOUT)  --  ゴーストにつかまった
        return
      end
    else
      a.laugh = false
      a.img = images.dai0
    end
    for i = 1, nfish + nghosts do
      if i > nfish then
        f = ghosts[i - nfish]
      else
        f = fish[i]
      end
      if f and a ~= f then
        r = collide(a, f, 32)
        if r >= 0 then
          dir = opposite(dir) -- 向きを変える
          break
        end
      end
    end
  else
    return -- 何もしない
  end
  if dir ~= a.direction then
    if a.kind == 0 then  -- 人
      dir = a.direction   -- 向きは変えない
      x, y = xsave, ysave -- 立ち止まる
    else
      x, y = movexy(xsave, ysave, dir, ds) -- 計算し直す
    end
  end
  a.direction = dir
  a.x, a.y = x, y
end

function do_game(dt)
  local i, dx, dy, dir
  local ctime = love.timer.getTime()
  if stime == nil then
    stime = ctime    --  開始時刻を記録
  end
  if mtime == nil or mtime < ctime then
    --  BGM の次のフレーズを演奏
    --  演奏するフレーズ：sounds.bgm[n], n = (mc < 42 ? 1 : (mc - 37) div 5)
    --  演奏時間：240 / (180 + (n - 1) * 8)秒
    local n = 1
    if mc >= 42 then n = math.floor((mc - 37) / 5) end
    if n > 12 then n = 12 end
    mtime = ctime + 240 / (180 + (n - 1) * 8)
    mc = mc + 1
    sounds.bgm[n]:stop()
    sounds.bgm[n]:play()
  end
  rem_time = stage_time - math.floor(ctime - stime)
  if rem_time < 0 then
    -- 時間切れ
    rem_time = 0
    change_status(S_FADEOUT)
    return
  end
  for i = 1, nfish do
    proceed(fish[i], dt)
  end
  for i = 1, nghosts do
    if ghosts[i] then proceed(ghosts[i], dt) end
  end
  dx, dy = 1, 1
  if love.keyboard.isDown("right") then
    dx = dx + 1
  end
  if love.keyboard.isDown("left") then
    dx = dx - 1
  end
  if love.keyboard.isDown("down")  then
    dy = dy + 1
  end
  if love.keyboard.isDown("up") then
    dy = dy - 1
  end
  dir = rdirect[dx * 3 + dy + 1]
  if dir >= 0 then
    man.direction = dir
    proceed(man, dt)
    if status ~= S_GAME then return end  -- 途中でクリア、またはやられた
    man.direction = dir -- 何かにぶつかっても方向は変えない
  end
end

function do_fadeout(dt)  -- ゴーストにやられた：１秒かけて人をフェードアウト
  if stime == nil then
    stime = love.timer.getTime()  --  開始時刻を記録
    man.dead = true
    love.audio.stop()         --  BGMを全部止める
    sounds.dead:play()        --  やられた音を鳴らす
  end
  man.ctime = love.timer.getTime() - stime
  if man.ctime > 2 then  -- ２秒たったら
    nlives = nlives - 1  -- 残りキャラ数を１減らす
    man.ctime = nil
    if nlives == 0 then
      change_status(S_OVER)   -- ゲームオーバー
    else
      change_status(S_NEWSTAGE) -- このステージやり直し
    end
  end
end

function do_title(dt)
  local i
  if stime == nil then
    stime = love.timer.getTime()  --  開始時刻を記録
    nfish = 6
    fish = {}
    man = nil
    setup_fish()
    nghosts = 0
    sounds.title:play()
  end
  for i = 1, nfish do
    proceed(fish[i], dt)
  end
  if love.keyboard.isDown("space") then
    change_status(S_NEWGAME)
    sounds.title:stop()
  end
end

function do_attraction(dt)
  local dx = (width - 64) / 2 / 18
  local n, x, dir, dir2, i
  if stime == nil then
    stime = love.timer.getTime()  --  開始時刻を記録
    count = -1
    nfish = 8
    for i = 1, nfish do
      fish[i].dead = false
      fish[i].img = images.fish0
      fish[i].y = 160 + 100 * (i % 2)
    end
    man.x = math.floor(width / 2)
    man.y = 210
    man.direction = 0
    nghosts = 0
    sounds.pomp:play()
  end
  local ncount = math.floor((love.timer.getTime() - stime) / 0.25)
  if ncount == count then return end
  count = ncount
  if count < 20 then
    n = math.floor(count / 2)
    x = dx * n
    dir = 0
    if n == 3 or n == 7 then
      dir = 3
    end
  elseif count < 24 then
    x = dx * (10 + count - 20)
    if count == 20 or count == 22 then
      dir = 0
    elseif count == 21 then
      dir = 1
    else
      dir = 7
    end
  elseif count < 28 then
    x = dx * (14 + math.floor((count - 24) / 2))
    if count < 26 then
      dir = 0
    else
      dir = 3
    end
  elseif count < 36 then
    x = dx * (16 + math.floor((count - 28) / 4))
    if count < 32 then
      dir = 0
    else
      dir = 7
    end
  else
    stage = stage + 1
    change_status(S_NEWSTAGE)
    return
  end
  dir2 = (12 - dir) % 8
  x = x + 16
  fish[1].x = x
  fish[2].x = x
  fish[3].x = x - 48
  fish[4].x = x - 48
  fish[5].x = width - x
  fish[6].x = fish[5].x
  fish[7].x = fish[5].x + 48
  fish[8].x = fish[7].x
  for i = 1, 4 do
    fish[i].direction = dir
    fish[i + 4].direction = dir2
  end
end

function do_gameover(dt)
  change_status(S_TITLE)
  msg_height = 0
end

function love.update(dt)
  if status == S_TITLE then
    -- タイトル画面
    do_title(dt)
  elseif status == S_OVER then
    -- ゲームオーバー画面
    do_gameover(dt)
  elseif status == S_SHOW then
    -- アトラクション
    do_attraction(dt)
  elseif status == S_NEWGAME or status == S_NEWSTAGE then
    if status == S_NEWGAME then
      -- ゲームを初期化
      stage = 1
      score = 0
      nlives = 3
    end
    -- ステージを初期化
    msg_height = 16
    prepare_stage()
    draw_present()
    sounds.start:play()   -- ファンファーレを鳴らす
    wait_duration(1.5)    -- 1.5秒待つ
    change_status(S_GAME)
  elseif status == S_FADEOUT then
    do_fadeout(dt)
  else
    -- 普通のゲーム処理
    do_game(dt)
  end
end

function drawobj(a)
  local kind = a.kind
  local d, sy = 0, 1
  if kind < 2 then  -- 人、さかな：進行方向を向く
    d = a.direction
    if d >= 2 and d < 6 then sy = -1 end
  end
  love.graphics.draw(a.img, a.x, a.y, 0.7854 * d, -1, sy, a.hw, a.hw)
end

-- コールバック以外で画面を書き換える
function draw_present()
  love.graphics.clear(love.graphics.getBackgroundColor())
  love.graphics.origin()
  love.draw()
  love.graphics.present()
end

-- ある時間待つ
function wait_duration(duration)
  love.timer.sleep(duration)
  love.timer.step()  -- 今待った時間を次の update に含めない
end

function love.draw()
  local i
  love.graphics.setColor(100, 100, 100)
  if transy > 0 then
    love.graphics.rectangle("fill", 0, 0, swidth, transy)
    love.graphics.rectangle("fill", 0, sheight - transy, swidth, transy)
  end
  if transx > 0 then
    love.graphics.rectangle("fill", 0, transy, transx, sheight - 2 * transy)
    love.graphics.rectangle("fill", swidth - transx, transy, transx, sheight - 2 * transy)
  end
  love.graphics.setColor(255, 255, 255)
  love.graphics.translate(transx, transy)
  love.graphics.scale(scale, scale)
  love.graphics.setScissor(transx, transy, swidth - transx * 2, sheight - transy * 2)
  for i = 1, nfish do
    if fish[i].alive then
      drawobj(fish[i])
    end
  end
  for i = 1, nghosts do
    local g = ghosts[i]
    if g and not g.hidden then
      drawobj(g)
    end
  end
  if man then
    if man.ctime then
      local c = math.floor(255 * (1 - man.ctime))
      if c < 0 then c = 0 end
      love.graphics.setColor(c, 0, 0)
    end
    drawobj(man)
  end
  if status == S_TITLE then
    love.graphics.draw(images.title1, 80, 54, 0, 1, 1)
    love.graphics.draw(images.title2, 155, 144, 0, 1, 1)
    love.graphics.draw(images.title4, 146, 270, 0, 1, 1)
  end
  if status == S_SHOW then
    love.graphics.draw(images.clear1, 125, 54, 0, 1, 1)
  end
  if msg_height > 0 then
    love.graphics.setFont(font)
    if rem_time < 10 then
      love.graphics.setColor(255, 0, 0)
      love.graphics.rectangle("fill", 24, 0, 80, 16)
      love.graphics.setColor(255, 255, 255)
    elseif rem_time < 30 then
      love.graphics.setColor(255, 255, 0, 176)
      love.graphics.rectangle("fill", 24, 0, 80, 16)
      love.graphics.setColor(0, 100, 0)
    else
      love.graphics.setColor(255, 255, 255)
    end
    love.graphics.print(string.format("Time %3d", rem_time), 32, 0)
    love.graphics.setColor(255, 255, 255)
    love.graphics.print(string.format("Scene %2d    Score %06d    Life %d", stage, score, nlives), 128, 0)
  end
  love.graphics.setScissor()
end
